/*------------------------------------------------
--                 LAZZE-GETOBZ                 --
--------------------------------------------------
-- Filename      : network.c
-- Description   : The network interface 
--                 
--
-- Author        : Lars Englund
--                 d99las@dtek.chalmers.se
--
--                 Tobias Gedell
--                 d99getob@dtek.chalmers.se
--
-- Created       : 020831
-- Last update   : 020902
------------------------------------------------*/

#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>


#include "error.h"
#include "network.h"
#include "game.h"
#include "map.h"
#include "robot.h"

int serverSocket;

char *dispDir(int dir);

void sendString(char *str);
char *readLine();

char lineBuffer[100000];

void connectToServer(char *host, int port)
{
    struct sockaddr_in addr;
    struct hostent* pHostInfo;
    
    
    //- Get host name
    if((pHostInfo=gethostbyname(host))==NULL)
        ERROR("Invalid host!");
    
    //- Create socket
    if((serverSocket=socket(AF_INET, SOCK_STREAM, 0))==-1)
        ERROR("Cannot create socket");

    //- Setup address
    addr.sin_family=AF_INET;
    addr.sin_port=htons(port);
    memcpy(&addr.sin_addr.s_addr, pHostInfo->h_addr_list[0], pHostInfo->h_length);
    
    //- Connect
    if(connect(serverSocket, (struct sockaddr*)&addr, sizeof(addr))==-1)
        ERROR("Cannot connect to server");
        
    #ifdef DEBUG
    printf("Connected to server\n");
    #endif
}

void sendInit()
{
    char *line;
    
    int boardWidth, boardHeight;
    int id, carry, money;
    int ggr, ggr2;

    #ifdef DEBUG
    printf("Sending init\n");
    #endif

    //- Send player
    sendString("Player\n");
    
    //- Receive board
    line=readLine();
    
    sscanf(line, "%d %d", &boardWidth, &boardHeight);
    
    #ifdef DEBUG
    printf("boardWidth=%d\n", boardWidth);
    printf("boardHeight=%d\n", boardHeight);
    #endif
    
    //- Init maps
    allocMaps(boardWidth, boardHeight);

    //- Read board lines
    for(ggr=0;ggr<boardHeight;ggr++)
    {
        line=readLine();
        
        for(ggr2=0;ggr2<boardWidth;ggr2++)
        {
            putGamePoint(ggr2, ggr, line[ggr2]);
        }
    }
    
    //- Receive player configuration
    line=readLine();
    sscanf(line, "%d %d %d", &id, &carry, &money);
    
    #ifdef DEBUG
    printf("id=%d\n", id);
    printf("carry=%d\n", carry);
    printf("money=%d\n", money);
    #endif
    
    robotId=id;
    robotCapacity=carry;
    robotMoney=money;

    //- Receive initial server response
    getReply();
}

void sendMove(int bid, int dir)
{
    char buffer[1000];
    
    sprintf(buffer, "%d Move %s\n", bid, dispDir(dir));
    
    sendString(buffer);
}

void sendPick(int bid, int *ids)
{
    int ggr;

    char buffer[1000];
    

    sprintf(buffer, "%d Pick", bid);
    
    #ifdef DEBUG
    printf("Sending command: %s", buffer);
    #endif

    sendString(buffer);
    
    ggr=0;
    while(ids[ggr]!=-1)
    {
        sprintf(buffer, " %d", ids[ggr]);
        
        #ifdef DEBUG
        printf("%s", buffer);
        #endif

        sendString(buffer);
        
        ggr++;
    }
    
    #ifdef DEBUG
    printf("\n");
    #endif
    
    sendString("\n");
}

void sendDrop(int bid, int *ids)
{
    int ggr;

    char buffer[1000];
    
    sprintf(buffer, "%d Drop", bid);
    
    #ifdef DEBUG
    printf("Sending command: %s", buffer);
    #endif

    sendString(buffer);

    ggr=0;
    while(ids[ggr]!=-1)
    {
        sprintf(buffer, " %d", ids[ggr]);
        
        #ifdef DEBUG
        printf("%s", buffer);
        #endif

        sendString(buffer);
        
        ggr++;
    }
    
    #ifdef DEBUG
    printf("\n");
    #endif
    
    sendString("\n");
}

void getReply()
{
    char *line;
    char *ptr;
    
    int id=-1;
    
    line=readLine();
    #ifdef DEBUG
    printf("Reply: %s\n", line);    
    #endif
    
    ptr=strtok(line, " ");
    
    while(ptr!=NULL)
    {
        if(ptr[0]!='#'&&id==-1)
        {
            ERROR("No '#' in getReply, no previous player");
        }
        else if(ptr[0]!='#')
        {
            #ifdef DEBUG
//            printf("Player %d (continued)\n", id);
            #endif
        }
        else
        {
            id=atoi(&ptr[1]);

            #ifdef DEBUG
//            printf("Player %d\n", id);
            #endif
        
            //- Update timestamp for player
            timeStampPlayer(id);

            if((ptr=strtok(NULL, " "))==NULL) break;
        }
        
        
        /* We can get a new player id here */
        if(ptr[0]=='#') continue;

        if(strcmp(ptr, "N")==0)
        {
            #ifdef DEBUG
//            printf("Player %d moved N\n", id);
            #endif

            movePlayer(id, UP);
        }
        else if(strcmp(ptr, "W")==0)
        {
            #ifdef DEBUG
//            printf("Player %d moved W\n", id);
            #endif

            movePlayer(id, LEFT);
        }
        else if(strcmp(ptr, "E")==0)
        {
            #ifdef DEBUG
//            printf("Player %d moved E\n", id);
            #endif

            movePlayer(id, RIGHT);
        }
        else if(strcmp(ptr, "S")==0)
        {
            #ifdef DEBUG
//            printf("Player %d moved S\n", id);
            #endif

            movePlayer(id, DOWN);
        }
        else if(strcmp(ptr, "P")==0)
        {
            int pid;
            
            if((ptr=strtok(NULL, " "))==NULL)
                ERROR("No token after P in getReply");
                
            pid=atoi(ptr);
    
            #ifdef DEBUG
            printf("Player %d picked %d\n", id, pid);
            #endif

            playerPick(id, pid);
        }
        else if(strcmp(ptr, "D")==0)
        {
            int pid;
            
            if((ptr=strtok(NULL, " "))==NULL)
                ERROR("No token after D in getReply");
                
            pid=atoi(ptr);
    
            #ifdef DEBUG
            printf("Player %d dropped %d\n", id, pid);
            #endif

            playerDrop(id, pid);
        }
        else if(strcmp(ptr, "X")==0)
        {
            int x, y;
            int pid;
            
            if((ptr=strtok(NULL, " "))==NULL)
                ERROR("No token after X in getReply");
                
            x=atoi(ptr)-1;
    
            if((ptr=strtok(NULL, " "))==NULL)
                ERROR("No token after X pos in getReply");
                
            if(strcmp(ptr, "Y")!=0)
                ERROR("No Y found after X pos in getReply");

            if((ptr=strtok(NULL, " "))==NULL)
                ERROR("No token after Y in getReply");
                
            y=atoi(ptr)-1;

            #ifdef DEBUG
            printf("Player %d appeared at (%d, %d)\n", id, x, y);
            #endif

            addPlayer(x, y, id);
        }
        else
        {
            ERROR("Invalid token after id in getReply");
        }
        
        ptr=strtok(NULL, " ");
    }
}

void getPackets()
{
    char *line;
    char *ptr;
    
    line=readLine();
    #ifdef DEBUG
//    printf("Packets: %s\n", line);    
    #endif
    
    ptr=strtok(line, " ");
    while(ptr!=NULL)
    {
        int id;
        int x, y;
        int weight;

        id=atoi(ptr);

        if((ptr=strtok(NULL, " "))==NULL)
            ERROR("No token x in getPackets");
        x=atoi(ptr)-1;

        if((ptr=strtok(NULL, " "))==NULL)
            ERROR("No token y in getPackets");
        y=atoi(ptr)-1;

        if((ptr=strtok(NULL, " "))==NULL)
            ERROR("No token weight in getPackets");
        weight=atoi(ptr);

        #ifdef DEBUG
//        printf("Packet %d to (%d, %d) weight %d\n", id, x, y, weight);
        #endif
        
        addPacket(id, x, y, weight);
        
        ptr=strtok(NULL, " ");
    }    
}

void disconnectFromServer()
{
    #ifdef DEBUG
    printf("Disconnecting from server\n");
    #endif

    close(serverSocket);
}


char *dispDir(int dir)
{
    static char strUp[]="N";
    static char strDown[]="S";
    static char strLeft[]="W";
    static char strRight[]="E";
    
    switch(dir)
    {
        case UP:
            return strUp;

        case DOWN:
            return strDown;

        case LEFT:
            return strLeft;

        case RIGHT:
            return strRight;
    }
    ERROR("Invalid direction");    
}

void sendString(char *str)
{
    int left=strlen(str);
    int written;

    #ifdef DEBUG
//    printf("Sending string \"%s\"\n", str);
    #endif

    while(left>0)
    {
        written=write(serverSocket, str, strlen(str));
        
        if(written==-1) ERROR("written=-1");
        
        str+=written;
        left-=written;
    }
    
    #ifdef DEBUG
//    printf("String sent.\n");
    #endif
}

char *readLine()
{
    char buffer[10000];
    
    int count=0;
    int stop=0;
    
    char readChar;
    
    //- Read line
    while(stop==0)
    {
        switch(read(serverSocket, buffer, 1))
        {
            case 0:
                #ifdef DEBUG
                printf("read returned 0!\n");
                #endif
                break;
                
            case 1:
                readChar=buffer[0];
                lineBuffer[count]=readChar;
                
                #ifdef DEBUG
//                printf("('%c', %d)", readChar);
                #endif

                if(readChar=='\n')
                {
                    lineBuffer[count]='\0';
                    stop=1;
                }
                count++;
                break;
                
            default:
                ERROR("Invalid result from read");
        }
    }
    
    return lineBuffer;
}
